/* ml_glframe.c     Wrappers for framebuffer object extension
 *
 * http://khronos.org/registry/gles/extensions/OES/OES_framebuffer_object.txt
 */
#ifdef _WIN32
#include <wtypes.h>
#endif
#include <string.h>
#include <caml/misc.h>
#include <caml/mlvalues.h>
#include <caml/memory.h>
#include <caml/alloc.h>
#include <caml/config.h>
#ifdef __APPLE__
#include <OpenGLES/ES1/gl.h>
#include <OpenGLES/ES1/glext.h>
#else
#include <GLES/gl.h>
#include <GLES/glext.h>
#endif
#include "ml_gl.h"
#include "gl_tags.h"

#if GL_OES_framebuffer_object

ML_1_ (glIsRenderbufferOES, (GLuint) Int32_val, Val_bool)
ML_2 (glBindRenderbufferOES, GLenum_val, (GLuint) Int32_val)

CAMLprim value ml_glDeleteRenderbuffer(value rbid)
{
    GLuint id = Int32_val(rbid);
    glDeleteRenderbuffersOES(1, &id);
    return Val_unit;
}

CAMLprim value ml_glGenRenderbuffer(value unit)
{
    GLuint id;
    glGenRenderbuffersOES(1, &id);
    return copy_int32(id);
}

CAMLprim value ml_glRenderbufferStorageOES(value target, value internal,
                        value width, value height) 
{
    /* We translate the internal format by hand because some values are
     * optional (i.e., not defined in every implementation).
     */
    GLenum i;
    switch(internal) {
#if GL_OES_rgb8_rgba8
    case MLTAG_rgba8: i = GL_RGBA8_OES; break;
    case MLTAG_rgb8: i = GL_RGB8_OES; break;
#endif
#if GL_OES_depth24
    case MLTAG_depth_component24: i = GL_DEPTH_COMPONENT24_OES; break;
#endif
#if GL_OES_depth32
    case MLTAG_depth_component32: i = GL_DEPTH_COMPONENT32_OES; break;
#endif
#if GL_OES_stencil1
    case MLTAG_stencil_index1: i = GL_STENCIL_INDEX1_OES; break;
#endif
#if GL_OES_stencil4
    case MLTAG_stencil_index4: i = GL_STENCIL_INDEX4_OES; break;
#endif
#if GL_OES_stencil8
    case MLTAG_stencil_index8: i = GL_STENCIL_INDEX8_OES; break;
#endif
    /* The default case catches all the required internal formats.
     */
    default: i = GLenum_val(internal); break;
    }
    glRenderbufferStorageOES(GLenum_val(target), i, Int_val(width),
                                Int_val(height));
    return Val_unit;
}

CAMLprim value ml_glGetRenderbufferParameter(value target, value param)
{
    /* The parameter returns a single GLint value.
     *
     * We translate by hand to renderbuffer enums because the names are
     * generic.  (E.g., width should correspond to GL_WIDTH.)
     */
    GLint i;
    GLenum p;
    switch(param) {
    case MLTAG_width: p = GL_RENDERBUFFER_WIDTH_OES; break;
    case MLTAG_height: p = GL_RENDERBUFFER_HEIGHT_OES; break;
    case MLTAG_red_size: p = GL_RENDERBUFFER_RED_SIZE_OES; break;
    case MLTAG_green_size: p = GL_RENDERBUFFER_GREEN_SIZE_OES; break;
    case MLTAG_blue_size: p = GL_RENDERBUFFER_BLUE_SIZE_OES; break;
    case MLTAG_alpha_size: p = GL_RENDERBUFFER_ALPHA_SIZE_OES; break;
    case MLTAG_depth_size: p = GL_RENDERBUFFER_DEPTH_SIZE_OES; break;
    case MLTAG_stencil_size: p = GL_RENDERBUFFER_STENCIL_SIZE_OES; break;
    default: ml_raise_gl("glGetRenderbufferParameter: unknown param");
    }
    glGetRenderbufferParameterivOES(GLenum_val(target), p, &i);
    return Val_int(i);
}

CAMLprim value ml_glGetRenderbufferParameter_fmt(value target, value param)
{
    /* The parameter returns a renderbuffer internal format.
     */
    GLint i;
    GLenum p;
    switch(param) {
    case MLTAG_internal_format: p = GL_RENDERBUFFER_INTERNAL_FORMAT_OES; break;
    default: ml_raise_gl("glGetRenderbufferParameter: unknown param");
    }
    glGetRenderbufferParameterivOES(GLenum_val(target), p, &i);
    switch(i) {
    case GL_RGB565_OES: return MLTAG_rgb565;
    case GL_RGBA4_OES: return MLTAG_rgba4;
    case GL_RGB5_A1_OES: return MLTAG_rgb5_a1;
    case GL_DEPTH_COMPONENT16_OES: return MLTAG_depth_component16;

#if GL_OES_rgb8_rgba8
    case GL_RGBA8_OES: return MLTAG_rgba8;
    case GL_RGB8_OES: return MLTAG_rgb8;
#endif
#if GL_OES_depth24
    case GL_DEPTH_COMPONENT24_OES: return MLTAG_depth_component24;
#endif
#if GL_OES_depth32
    case GL_DEPTH_COMPONENT32_OES: return MLTAG_depth_component32;
#endif
#if GL_OES_stencil1
    case GL_STENCIL_INDEX1_OES: return MLTAG_stencil_index1;
#endif
#if GL_OES_stencil4
    case GL_STENCIL_INDEX4_OES: return MLTAG_stencil_index4;
#endif
#if GL_OES_stencil8
    case GL_STENCIL_INDEX8_OES: return MLTAG_stencil_index8;
#endif
    }
    ml_raise_gl("glGetRenderbufferParameter: unknown internal format");
}

ML_1_ (glIsFramebufferOES, (GLuint) Int32_val, Val_bool)
ML_2 (glBindFramebufferOES, GLenum_val, (GLuint) Int32_val)

CAMLprim value ml_glDeleteFramebuffer(value fbid)
{
    GLuint id = Int32_val(fbid);
    glDeleteFramebuffersOES(1, &id);
    return Val_unit;
}

CAMLprim value ml_glGenFramebuffer(value unit)
{
    GLuint id;
    glGenFramebuffersOES(1, &id);
    return copy_int32(id);
}

CAMLprim value ml_glCheckFramebufferStatus(value target)
{
    GLenum status = glCheckFramebufferStatusOES(GLenum_val(target));
    switch(status) {
    case GL_FRAMEBUFFER_COMPLETE_OES:
        return MLTAG_complete;
    case GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_OES:
        return MLTAG_incomplete_attachment;
    case GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_OES:
        return MLTAG_incomplete_missing_attachment;
    case GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_OES:
        return MLTAG_incomplete_dimensions;
    case GL_FRAMEBUFFER_INCOMPLETE_FORMATS_OES:
        return MLTAG_incomplete_formats;
    case GL_FRAMEBUFFER_UNSUPPORTED_OES:
        return MLTAG_unsupported;
    }
    ml_raise_gl("glCheckFramebufferStatus: unknown status");
}

CAMLprim value ml_glFramebufferRenderbuffer(value target, value attachment,
                    value rendertarget, value rbid)
{
    GLenum att;
    
    switch(attachment) {
    case MLTAG_color0: att = GL_COLOR_ATTACHMENT0_OES; break;
    case MLTAG_depth: att = GL_DEPTH_ATTACHMENT_OES; break;
    case MLTAG_stencil: att = GL_STENCIL_ATTACHMENT_OES; break;
    default: ml_raise_gl("glFramebufferRenderbuffer: unknown attachment");
    }
    glFramebufferRenderbufferOES(GLenum_val(target), att,
                                GLenum_val(rendertarget), Int32_val(rbid));
    return Val_unit;
}

CAMLprim value ml_glFramebufferTexture2D(value target, value attachment,
                    value textarget, value texid, value level)
{
    GLenum att;
    
    switch(attachment) {
    case MLTAG_color0: att = GL_COLOR_ATTACHMENT0_OES; break;
    case MLTAG_depth: att = GL_DEPTH_ATTACHMENT_OES; break;
    case MLTAG_stencil: att = GL_STENCIL_ATTACHMENT_OES; break;
    default: ml_raise_gl("glFramebufferTexture2D: unknown attachment");
    }
    glFramebufferTexture2DOES(GLenum_val(target), att, GLenum_val(textarget),
                                Int32_val(texid), Int_val(level));
    return Val_unit;
}

#ifndef GL_NONE_OES
#define GL_NONE_OES 0
#endif

CAMLprim value ml_glGetAttachmentType(value target, value attachment)
{
    CAMLparam2(target, attachment);
    GLenum targ = GLenum_val(target);
    CAMLlocal1(result);
    GLenum a;
    GLint i;
    switch(attachment) {
    case MLTAG_color0: a = GL_COLOR_ATTACHMENT0_OES; break;
    case MLTAG_depth: a = GL_DEPTH_ATTACHMENT_OES; break;
    case MLTAG_stencil: a = GL_STENCIL_ATTACHMENT_OES; break;
    default:
        ml_raise_gl("glGetFramebufferAttachmentType: unknown attachment type");
        break;
    }
    i = 0;
    glGetFramebufferAttachmentParameterivOES(targ, a,
            GL_FRAMEBUFFER_ATTACHMENT_OBJECT_TYPE_OES, &i);
    switch(i) {
    case GL_NONE_OES:
        result = MLTAG_none;
        break;
    case GL_TEXTURE:
    case GL_RENDERBUFFER_OES:
        result = caml_alloc(2, 0);
        Store_field(result, 0,
                i == GL_TEXTURE ? MLTAG_texture : MLTAG_renderbuffer);
        i = 0;
        glGetFramebufferAttachmentParameterivOES(targ, a,
                GL_FRAMEBUFFER_ATTACHMENT_OBJECT_NAME_OES, &i);
        Store_field(result, 1, caml_copy_int32(i));
        break;
    default:
        ml_raise_gl("glGetFramebufferAttachmentType: unknown object type");
        break;
    }
    CAMLreturn(result);
}

ML_1 (glGenerateMipmapOES, GLenum_val)

CAMLprim value ml_glTranslateEnum(value glenum)
{
    return Val_int(GLenum_val(glenum));
}

#else /* !GL_OES_framebuffer_object */

/* Framebuffer object functions fail at runtime if not supported.
 */

CAMLprim value ml_glIsRenderbufferOES(value arg1)
{
    ml_raise_gl("Function: glIsRenderbufferOES not available");
}

CAMLprim value ml_glBindRenderbufferOES(value arg1, value arg2)
{
    ml_raise_gl("Function: glBindRenderbufferOES not available");
}

CAMLprim value ml_glDeleteRenderbuffer(value rbid)
{
    ml_raise_gl("Function: glDeleteRenderbuffersOES not available");
}

CAMLprim value ml_glGenRenderbuffer(value unit)
{
    ml_raise_gl("Function: glGenRenderbuffersOES not available");
}

CAMLprim value ml_glRenderbufferStorageOES(value arg1, value arg2,
                        value arg3, value arg4) 
{
    ml_raise_gl("Function: glRenderbufferStorageOES not available");
}

CAMLprim value ml_glGetRenderbufferParameter(value target, value param)
{
    ml_raise_gl("Function: glGetRenderbufferParameterivOES not available");
}

CAMLprim value ml_glGetRenderbufferParameter_fmt(value target, value param)
{
    ml_raise_gl("Function: glGetRenderbufferParameterivOES not available");
}

CAMLprim value ml_glIsFramebufferOES(value arg1)
{
    ml_raise_gl("Function: glIsFramebufferOES not available");
}

CAMLprim value ml_glBindFramebufferOES(value arg1, value arg2)
{
    ml_raise_gl("Function: glBindFramebufferOES not available");
}

CAMLprim value ml_glDeleteFramebuffer(value fbid)
{
    ml_raise_gl("Function: glDeleteFramebuffersOES not available");
}

CAMLprim value ml_glGenFramebuffer(value unit)
{
    ml_raise_gl("Function: glGenFramebuffersOES not available");
}

CAMLprim value ml_glCheckFramebufferStatus(value target)
{
    ml_raise_gl("Function: glCheckFramebufferStatusOES not available");
}

CAMLprim value ml_glFramebufferRenderbuffer(value target, value attachment,
                    value rendertarget, value rbid)
{
    ml_raise_gl("Function: glFramebufferRenderbufferOES not available");
}

CAMLprim value ml_glFramebufferTexture2D(value target, value attachment,
                    value textarget, value texid, value level)
{
    ml_raise_gl("Function: glFramebufferTexture2DOES not available");
}

CAMLprim value ml_glGetAttachmentType(value target, value attachment)
{
    ml_raise_gl("Function: glGetFramebufferAttachmentParameterivOES not available");
}

CAMLprim value ml_glGenerateMipmapOES(value arg1)
{
    ml_raise_gl("Function: glGenerateMipmapOES not available");
}

CAMLprim value ml_glTranslateEnum(value glenum)
{
    /* This doesn't actually require any OpenGL functionality at all,
     * but make it undefined for consistency.
     */
    ml_raise_gl("Function: glTranslateEnum not available");
}
#endif /* !GL_OES_framebuffer_object */
